function dag = learn_struct_tan(data, class_node, root, node_sizes, scoring_fn)
% LEARN_STRUCT_TAN Learn the structure of the tree augmented naive bayesian network 
% (with discrete nodes)
% dag = learn_struct_tan(app, class, root)
%
% Input :
% 	data(i,m) is the value of node i in case m
% 	class_node is the class node
% 	root is the root node of the tree part of the dag (must be different from the class node)
%   	node_sizes = 1 if gaussian node,
%   	scoring_fn = 'bic' (default value) or 'mutual_info'
%
% Output :
%	dag = adjacency matrix of the dag
%
% V1.1 : 21 may 2003, (O. Francois, Ph. Leray)
% V1.2 : may 2005 bug correction about node types (Navid Serrano <Navid.Serrano@jpl.nasa.gov>)
% V1.3 : may 2005 bug correction score (PhL)


if nargin <4
    error('Requires at least 4 arguments.')
end

if nargin == 4
    scoring_fn='bic';
end;

if class_node==root
    error(' The root node can''t be the class node.');
end

if root>class_node
    root=root-1;
end

N=size(data,1);
discrete=double(not(node_sizes==1));

node_types=cell(N-1,1);
XC=setdiff(1:N,class_node);

for i=1:N-1
    if node_sizes(XC(i))==1
        node_types{i}='gaussian';
    else
        node_types{i}='tabular';
    end
end



score_mat=zeros(N-1,N-1);

switch scoring_fn
    case 'bic',
        for i=1:(N-2)
            score2 = score_family(XC(i), [class_node], 'tabular', 'bic', node_sizes, discrete, data,[]);
            for j=(i+1):(N-1)
                score1 = score_family(XC(i), [XC(j) class_node], node_types{i}, 'bic', node_sizes, discrete, data,[]);
                score = score2-score1;
                score_mat(i,j)=score;
                score_mat(j,i)=score;
            end
            
        end
    case 'mutual_info',
        for i=1:(N-2)
            for j=(i+1):(N-1)
                score_mat(i,j)= -cond_mutual_info_score(XC(i),node_sizes(XC(i)),XC(j),node_sizes(XC(j)),class_node,node_sizes(class_node),data);
                score_mat(j,i)=score_mat(i,j);
            end
        end
    otherwise,
        error(['unrecognized scoring fn ' scoring_fn]);
end

G = minimum_spanning_tree(score_mat);
T = mk_rooted_tree(G, root);
T=full(T);

dag=zeros(N,N);
dag(XC,XC)=T;
dag(class_node,XC)=1;
