function [bnet,lcl] = learn_params_elr(bnet,data,cnode,mode,gradfunc,bnet_initialized,thres,max_iter,try_mixim,line_search_func)

% LEARN_PARAMS_ELR Estimates params of a BN classifier for a fully observed model.
% [bnet,lcl] = learn_params_elr(bnet,data,cnode,mode,gradfunc,bnet_initialized,thres,max_iter,use_mixim,line_search_func)
%
% Performs parameters learning of a bayesian belief net classifier through
%  Extended Logistic Regression [Gre04].
% bnet is the bayesian belief net structure (in an programing sense).
% data is an [dim n] array, with n examples and dim variables.
% cnode is the decision node (default is last).
% Optional arguments max_iter and thresh are related to the gradient 
%  descent algorithm which can be 'scg' or 'conjgrad' (gradfunc).
% Possible values for mode are 'batch' (default) and 'inline' (warning: 
%  only 'batch' is available at the moment).
% use_mixim is a testing optimization.
% line_search_func should obey NetLab linemin function prototype.
%
% [Gre04] Greiner, R., Zhou, W., Su, X., Shen, B., Structural extension to 
%  logistic regression: Discriminative parameter learning of belief net 
%  classifiers, 2004.

if nargin<2
    error('too few input arguments')
end%if

[dim,n] = size(data);
epsilon = 1e-6;
cyclic = true;

if nargin<10 || length(line_search_mode)==0
    line_search_func = 'none';
end%if
if nargin<9 || length(use_mixim)==0
    use_mixim = false;
end%if
if nargin<8 || length(max_iter)==0
    max_iter = 1e2;
end%if
if nargin<7 || length(thres)==0
    thres = 1e-2;
end%if
if nargin<6 || length(bnet_initialized)==0
    bnet_initialized = false;
end%if
if nargin<5 || length(gradfunc)==0
    gradfunc = 'scg';
end%if
if nargin<4 || length(mode)==0
    mode = 'batch';
end%if
if nargin<3 || length(cnode)==0
    cnode = dim;
end%if

if ~strcmp(mode,'batch') && ~strcmp(mode,'inline')
    warning(['unknown mode ' mode '. switching to batch mode'])
    mode = 'batch';
end%if
if strcmp(mode,'inline')
    warning('inline mode quite intractable at the moment. you should think about switching to batch mode')
end%if
if exist(gradfunc,'file')~=2
    warning(['optimization function ' gradfunc ' does not exist. using scg instead'])
    gradfunc = 'scg';
end%if
if ~strcmp(line_search_func,'none') && exist(line_search_func,'file')~=2
    warning(['line search function ' line_search_func ' does not exist. skipping line search step'])
    line_search_func = 'none';
end%if

verbose = false; % hard coded at the moment (verbose=true is not very useful)

options = zeros(18,1);
if verbose
    options(1) = 1; % display error values (slows down the process a lot, not recommended)
end%if
% thres and max_iter are used in several ways; perhaps would it be better
% to have different values...
options(2) = thres; % absolute precision required for the value of X at the solution
options(3) = thres; % precision required of the objective function at the solution
%options(9) = 1; % check the user defined gradient function
options(14) = max_iter; % maximum number of iterations
options(15) = thres; % precision of the line search in parameter space (conjgrad)

enodes = [1:cnode-1 cnode+1:dim];

if ~bnet_initialized
    for i=1:dim
        bnet.CPD{i} = tabular_CPD(bnet,i);
    end%for
    bnet = learn_params(bnet,data); % parameters initialization
end%if
theta = bnet_unpak(bnet);

ind = find(theta<epsilon); % prevent logarithms of 0
theta(ind) = epsilon * ones(length(ind),1);
beta = log(theta)'; % softmax parameters initialization
% conjgrad or scg need that LCLgrad returns a row vector, therefore
% beta must be a row vector as well.

if use_mixim
    mixim = bnet_mixim(bnet,true);
else
    mixim = [];
end%if

% nested """anonymous""" functions to be evaluated by external functions (gradfunc).
% the m letter before LCL stands for 'minus' because we need a function to minimize.
mLCL =     @(x,varargin)  -LCL(x',bnet,data,cnode,enodes,cyclic,mixim);
mLCLgrad = @(x,examples,varargin)  -LCLgrad(x',bnet,examples,cnode,enodes,cyclic,mixim)';

if ~strcmp(line_search_func,'none')
    if verbose
        tic % disable use of tic/toc at a higher level (i. e., functions that call learn_params_elr cannot bracket this call with tic/toc)
    end%if
    
    % the following two values are computed because of the line search step
    mlcl = feval(mLCL,[]); % opposite of global log conditional likelyhood initialization
    mlclgradient = feval(mLCLgrad,[],data);

    % an error in linemin help postponed the end of learn_params_elr
    % development. the first output argument was described has the optimal
    % value of the parameters (beta*).

    % line search step
    [alpha,options] = feval(line_search_func,mLCL,beta,mlclgradient,mlcl,options);
    % actually, linemin is buggy (do not put the original linemin function 
    % at first in your Matlab path): alpha in [0,1] during call to minbrack
    % is a wrong choice for generic use.
    
    flog = options(8); % retrieve optimal value of the function
    beta = beta + alpha * mlclgradient;
    
    if verbose
        t = toc;
        disp(['line search step performed (elapsed time: ' num2str(t) 's)'])
    end%if
    
else
    flog = 1e10; % flog is used in inline mode
end%if


% the heavy task: gradient descent to find optimal softmax parameter values.
if strcmp(mode,'batch')
    if verbose
        tic
    end%if
    [beta,options,flog,ptlog] = feval(gradfunc,mLCL,beta,options,mLCLgrad,data);
    if verbose
        t = toc;
    end%if
else % strcmp(mode,'inline') % actually intractable as is
    t = 0; k = 0;
    flog = [+Inf flog]; % to have at least one loop (if max_iter>1)
    while flog(end-1)-flog(end)>thres && k<max_iter % perhaps would it be better to have different bounding values for line_search_func/gradfunc and those ones (thres and max_iter)...
        if verbose
            tic
        end%if
        for i=1:n % number of examples
            [beta,options,flog,ptlog] = feval(gradfunc,mLCL,beta,options,mLCLgrad,data(:,i));
        end%for
        if verbose
            t = t + toc;
            disp(['example ' num2str(i) ' out of ' num2str(n) ' (cumulated elapsed time: ' num2str(t) 's)'])
        end%if
        k = k + 1;
    end%while
end%if
if verbose
    disp(['gradient descent step performed (elapsed time: ' num2str(t) 's)'])
end%if


if use_mixim
    theta = beta2theta(mixim,beta');
else
    theta = beta2theta(bnet,beta');
end%if
bnet = bnet_pak(bnet,theta);

lcl = -flog(end);

end%function